# frozen_string_literal: true

require 'footrest/client'
require 'faraday'
require 'footrest'
require 'footrest/http_error'
require 'footrest/pagination'
require 'footrest/follow_redirects'
require 'footrest/parse_json'
require 'base64'
require 'paul_walker'

module BridgeAPI
  class Client < Footrest::Client
    require 'bridge_api/api_array'

    DATA_DUMP_DOWNLOAD_PATH = '/data_dumps/download'
    DATA_DUMP_PATH          = '/data_dumps'
    COURSE_TEMPLATE_PATH    = '/course_templates'
    ENROLLMENT_PATH         = '/enrollments'
    LTI_TOOLS_PATH          = '/lti_tools'
    PROGRAM_PATH            = '/programs'
    PROGRAM_ENROLLMENT_PATH = '/learners'
    USER_PATH               = '/users'
    GROUPS_PATH             = '/groups'
    MANAGER_PATH            = '/managers'
    ADMIN_PATH              = '/admin'
    AUTHOR_PATH             = '/author'
    LEARNER_PATH            = '/learner'
    LEARNERS_PATH           = '/learners'
    LEARNER_ITEMS_PATH      = '/learner_items'
    CUSTOM_FIELD_PATH       = '/custom_fields'
    SUB_ACCOUNT_PATH        = '/sub_accounts'
    SUPPORT_PATH            = '/support'
    ACCOUNT_PATH            = '/accounts'
    CLONE_OBJECTS_PATH      = '/clone_objects'
    API_VERSION             = 1
    API_PATH                = '/api'
    ROLE_PATH               = '/roles'
    AFFILIATED_SUBACCOUNTS  = '/affiliated_sub_accounts'
    BATCH_PATH              = '/batch'
    LIVE_COURSES_PATH       = '/live_courses'
    SESSIONS_PATH           = '/sessions'
    PUBLISH_PATH            = '/publish'
    WEB_CONFERENCE_PATH     = '/web_conference'
    RESTORE_PATH            = '/restore'
    DUE_DATE_PATH           = '/due_date'
    RESET_PATH              = '/reset'
    NEW_TEMPORARY_USERS     = '/new_temporary_users'
    LEARNER_ENROLLMENTS     = '/learner_enrollments'
    RESULT_MAPPING          = {}

    Dir[File.dirname(__FILE__) + '/client/*.rb'].each do |file|
      require file
      include const_get(File.basename(file).gsub('.rb', '').split('_').map(&:capitalize).join('').to_s)
    end

    def initialize(options = {}, &block)
      if BridgeAPI.enforce_rate_limits && has_token_pool?(options)
        options = initialize_from_token_pool(options)
      end
      super
    end

    # Override Footrest request for ApiArray support
    def request(method, &block)
      enforce_rate_limits if rate_limit_reached?
      response = connection.send(method, &block)
      apply_rate_limits(response)
      ApiArray.process_response(response, self, RESULT_MAPPING)
    end

    def has_token_pool?(config)
      config[:api_keys].is_a?(Hash) && config[:api_keys].keys.count >= 1 ||
        config[:api_tokens].is_a?(Array) && config[:api_tokens].count >= 1
    end

    def convert_tokens(config)
      return config unless config.has_key?(:api_tokens)
      return config unless config[:api_tokens].is_a?(Array)
      config[:api_keys] ||= {}
      config[:api_tokens].each do |token|
        decoded_token_array = Base64.strict_decode64(token).split(':')
        config[:api_keys][decoded_token_array[0]] = decoded_token_array[1]
      end
      config.delete(:api_tokens)
      config
    end

    def initialize_from_token_pool(config)
      config = convert_tokens(config)  
      creds = config[:api_keys].first
      config[:api_key]    ||= creds[0]
      config[:api_secret] ||= creds[1]
      config
    end

    # rotates to the next token in the pool (by order in which they were provided)
    def rotate_token! 
      return unless config[:api_keys].present?
      old_api_key = config[:api_key]
      keys = config[:api_keys].keys
      return if keys.count <= 1
      key                 = get_next_key(keys, config[:api_key])
      config[:api_key]    = key
      config[:api_secret] = config[:api_keys][key]
      set_connection(config)
      BridgeAPI.logger.debug("ROTATED TO KEY: #{config[:api_key]}")
    end

    def get_next_key(keys, current_key)
      keys.delete(current_key)
      usable_key = keys.find do |key|
        limit = rate_limit(key)
        current_key_limit = limit.present? ? limit.fetch('current') : 0
        BridgeAPI.beginning_rate_limit - current_key_limit > BridgeAPI.rate_limit_threshold
      end
      usable_key || keys[rand(keys.length)]
    end

    def rate_limit_reached?
      return false unless BridgeAPI.enforce_rate_limits && limit_remaining.present?

      limit_remaining < BridgeAPI.rate_limit_threshold
    end

    def enforce_rate_limits
      return unless rate_limit_reached?

      rotate_token! 
      if rate_limit_reached?
        sleep_time = rand(BridgeAPI.max_sleep_seconds)
        sleep_time = BridgeAPI.min_sleep_seconds if sleep_time < BridgeAPI.min_sleep_seconds 
        BridgeAPI.logger.debug("Rate limit reached sleeping for #{sleep_time}")
        sleep(sleep_time)
      end
    end

    def using_master_rate_limit?
      config[:master_rate_limit].present? || BridgeAPI.master_rate_limit.present?
    end

    def apply_rate_limits(response)
      limit = response.headers['x-rate-limit-remaining']
      return if limit.nil?

      BridgeAPI.logger.debug("BRIDGE RATE LIMIT REMAINING: #{limit} for key #{config[:api_key]}")
      self.limit_remaining = limit.to_i
    end

    def limit_remaining
      if using_master_rate_limit?
        limit = rate_limit(config[:api_key])
        if limit.nil?
          set_rate_limit(config[:api_key], 0)
          limit = { current: 0 }.with_indifferent_access
        end
        limit['current']
      else
        BridgeAPI.rate_limits[config[:api_key]]
      end
    end

    def limit_remaining=(value)
      if using_master_rate_limit?
        set_rate_limit(config[:api_key], value)
      else
        BridgeAPI.rate_limits[config[:api_key]] = value
      end
      refresh_stale_tokens
    end

    def set_rate_limit(key, limit)
      BridgeAPI.master_mutex.synchronize do
        PaulWalker::RateLimit.add(key, key, limit, BridgeAPI.beginning_rate_limit)
      end
    end

    def rate_limit(key)
      BridgeAPI.master_mutex.synchronize do
        PaulWalker::RateLimit.get(key, key)
      end
    end

    def refresh_stale_tokens
      return unless using_master_rate_limit?
      return unless config[:api_keys].present?
      api_keys = config[:api_keys].keys
      api_keys.delete(config[:api_key])
      api_keys.each do |key|
        limit = rate_limit(key)
        next unless limit.present?
        if limit['timestamp'].present? && DateTime.parse(limit['timestamp']) < 1.minute.ago
          BridgeAPI.logger.debug("Refreshing: #{key}")
          set_rate_limit(key, 0)
        end
      end
    end

    def set_connection(config)
      config[:logger] = config[:logging] if config[:logging]
      @connection     = Faraday.new(url: config[:prefix]) do |faraday|
        faraday.request :multipart
        faraday.request :url_encoded
        if config[:logger] == true
          faraday.response :logger
        elsif config[:logger]
          faraday.use Faraday::Response::Logger, config[:logger]
        end
        faraday.use Footrest::FollowRedirects, limit: 5 unless config[:follow_redirects] == false
        faraday.adapter Faraday.default_adapter
        faraday.use Footrest::ParseJson, content_type: /\bjson$/
        faraday.use Footrest::RaiseFootrestErrors
        faraday.use Footrest::Pagination
        faraday.headers[:accept]     = 'application/json'
        faraday.headers[:user_agent] = 'Footrest'
        if config[:api_key] && config[:api_secret]
          faraday.headers[:authorization] = 'Basic ' + Base64.strict_encode64("#{config[:api_key]}:#{config[:api_secret]}")
        elsif config[:token]
          faraday.headers[:authorization] = "Bearer #{config[:token]}"
        else
          raise 'No api authorization provided'
        end
      end
    end
  end
end
